         include      \masm32\include\kernel32.inc
         include      \masm32\include\masm32.inc

         includelib   \masm32\lib\kernel32.lib
         includelib   \masm32\lib\masm32.lib
;
         BmpFileTo24Bit PROTO :DWORD,:DWORD,:DWORD

         ColorIxPad     PROTO :DWORD,:DWORD

         Bitmapsize     PROTO :DWORD,:DWORD,:DWORD

; dougiem@dougiem.fsnet.co.uk
; http://www.dougiem.fsnet.co.uk/asm
.code
;________________________________________________________________________________
;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

; ************************************************************************
; * This routine decodes 1, 4, 8 and 24 .BMP files into 24 bit DIB data. *
; ************************************************************************
; eg.
;        invoke BmpFileTo24Bit,addr PathName,addr bmih,addr Pbuffer
;        jc ERROR
;        mov fsize,eax
;
;        *  code  *
;
;        invoke GlobalFree,Pbuffer
;
;
; Returns with your BIMAPINFOHEADER struture filled and lpBuffer is a pointer
; to the RGB pixel data

; in  -> lpszPathName
;     -> lpBmih   = pointer to a BITMAPINFOHEADER structure
;     -> lpBuffer = pointer to location to hold RGB address

; out <- lpBmih   = filled
;     <- lpBuffer = pointer filled with RGB pixel addr
;     <- carry flag set on error   eax = 0
;     <- carry flag clear          eax = file size

;***********************************************************************************
;*      NOTE: you must invoke GlobalFree,lpBuffer when finished with RGB buffer    *
;***********************************************************************************

BmpFileTo24Bit proc uses edi esi ecx ebx edx lpszPathName:dword,
                                                      lpBmih:dword,lpBuffer
local hBfile:dword      ; file handle
local Bsize :dword      ; bitmap size
local Pxsize:dword      ; size of pixel index
local Bfsize:dword      ; the file size
local Bread :dword      ; read bytes from file
local hBmem :dword      ; the bitmap pixel data addr
local hBmem2:dword      ; the pixel index
local hCol  :dword      ; color table addr
local Csize :dword      ; size of color table
local pad   :dword      ; number of dib pad bytes per line
local Cpad  :dword      ; pad bytes for color index
local BWa   :dword      ; one bit col value a
local BWb   :dword      ; one bit col value b
local BMPfh :BITMAPFILEHEADER
local BMPih :BITMAPINFOHEADER

; load data 
invoke CreateFile,lpszPathName,GENERIC_READ,FILE_SHARE_READ,NULL,
                           OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL
         cmp eax,INVALID_HANDLE_VALUE
         je ERROR
         mov hBfile,eax

         invoke GetFileSize,hBfile,0
         mov Bfsize,eax
     
     ; read in file header to local
         invoke ReadFile,hBfile,addr BMPfh,sizeof BITMAPFILEHEADER,ADDR Bread,NULL
     ; and info header to local
         invoke ReadFile,hBfile,addr BMPih,sizeof BITMAPINFOHEADER,ADDR Bread,NULL

     ; set up pixel data area for 24 bit bmp
         invoke Bitmapsize,BMPih.biWidth,BMPih.biHeight,addr pad
         mov Bsize,eax
         add eax,4 ; (for last byte reg overwrite)
         invoke GlobalAlloc,GPTR,eax
         mov hBmem,eax       ; addr of new bmp data

         mov eax,BMPfh.bfOffBits
         sub eax,sizeof BITMAPFILEHEADER
         sub eax,sizeof BITMAPINFOHEADER
         mov Csize,eax       ; size of color table

     .if eax != 0
     ; get color table into mem if col table
         invoke GlobalAlloc,GPTR,eax
         mov hCol,eax
         invoke ReadFile,hBfile,hCol,Csize,ADDR Bread,NULL

         mov eax,Bfsize
         sub eax,sizeof BITMAPFILEHEADER
         sub eax,sizeof BITMAPINFOHEADER
         sub eax,Csize
         mov Pxsize,eax      ; size of non 24 pix data

     ; and color info from pixel index
         invoke GlobalAlloc,GPTR,eax
         mov hBmem2,eax      ; addr of non 24 pix data
         invoke ReadFile,hBfile,hBmem2,Pxsize,ADDR Bread,NULL
         invoke CloseHandle,hBfile

     .else
     ; load 24 bit bmp pixel data into memory
         invoke ReadFile,hBfile,hBmem,Bsize,ADDR Bread,NULL
         invoke CloseHandle,hBfile
     .endif

         cmp BMPfh.bfType,"MB"    ; test its a bitmap
         jne ERROR

         movzx eax,BMPih.biBitCount

     .if eax == 1
         jmp OneBit
     .elseif eax == 4
         jmp FourBit
     .elseif eax == 8
         jmp EightBit
     .elseif eax == 24
         ; fall through to Bexit
     .else
         jmp ERROR
     .endif

  Bexit: invoke GlobalFree,hCol
         invoke GlobalFree,hBmem2

         mov eax,hBmem
         mov edi,lpBuffer
         mov [edi],eax            ; set callers pointer to pixel buffer

         mov word ptr BMPih.biBitCount,24
         mov BMPih.biCompression,BI_RGB 
         mov eax,Bsize
         mov BMPih.biSizeImage,eax 
         mov BMPih.biClrUsed,0
         mov BMPih.biClrImportant,0

         lea esi,BMPih
         mov edi,lpBmih
         mov ecx,[esi]
         rep movsb                ; fill callers info header

         mov eax,Bfsize           ; returns the file size
         clc
     ret

;________________________________________________________________________________

OneBit:  invoke ColorIxPad,BMPih.biWidth,1
         mov Cpad,eax   ; set the pad bytes for read color index
         mov edx,BMPih.biHeight   ; edx = height
         mov edi,hBmem  ; edi = ptr to final bitmap data
         mov esi,hBmem2 ; esi = ptr to 1 bit pixel data
         mov ebx,hCol   ; ptr to the color table
         mov eax,[ebx]  ; store the 2 dword colors
         mov BWa,eax    ; localy
         mov ebx,[ebx+4]
         mov BWb,ebx

  moreB: mov ecx,BMPih.biWidth    ; ecx = pixel count

  loopB: mov ebx,8      ; 8bit count
         mov al,[esi]   ; get 1 byte-> 8 pixels

  loopC: shl al,1       ; test bit in carry
         jc isone
         push eax
         mov eax,BWa
         mov [edi],eax  ; out 1st table dword
         pop eax
         jmp @F
  isone: push eax
         mov eax,BWb
         mov [edi],eax  ; out 2nd table value
         pop eax
     @@: add edi,3

         dec ecx        ; decrement width
         jne @F         ;-> end of line
         add edi,pad    ; add bmp pad bytes
         inc esi        ; point next
         add esi,Cpad   ; add color index pad
         dec edx        ; decrement height

         je Bexit       ; -> end
         jmp moreB      ; <- more lines

     @@: dec ebx        ; decrement eax bit count
         jne loopC      ; not 0 - do more shifts
         add esi,1      ; else new pix info
         jmp loopB      ; do more

;________________________________________________________________________________

FourBit: invoke ColorIxPad,BMPih.biWidth,4
         mov Cpad,eax
         mov edx,BMPih.biHeight
         mov edi,hBmem
         mov esi,hBmem2

  moreA: mov ecx,BMPih.biWidth
  loopA: mov eax,0
         mov al,[esi]
         push eax
         shr al,4       ; FIRST 4 BITS
         shl eax,2
         mov ebx,hCol   ; into table
         add ebx,eax
         mov eax,[ebx]  ; dword from table
         mov [edi],eax  ; into bitmap
         pop eax
         add edi,3
         dec ecx
         jne noend      ; until line finished

         inc esi
         jmp @F

  noend: and al,0fh     ; SECOND 4 BITS
         shl eax,2
         mov ebx,hCol   ; into table
         add ebx,eax
         mov eax,[ebx]  ; dword from table
         mov [edi],eax  ; into bitmap
         add edi,3
         inc esi        ; next index
         dec ecx
         jne loopA

     @@: add edi,pad    ; add pad bytes to bitmap area
         add esi,Cpad   ; pad bytes to index area
         dec edx
         jne moreA      ; new scan line

         jmp Bexit      ; end

;________________________________________________________________________________

EightBit: 
         invoke ColorIxPad,BMPih.biWidth,8
         mov Cpad,eax
         mov edx,BMPih.biHeight
         mov edi,hBmem
         mov esi,hBmem2

   more: mov ecx,BMPih.biWidth
     @@: mov eax,0
         mov al,[esi]   ; get 8bit index
         shl eax,2
         mov ebx,hCol   ; into table
         add ebx,eax
         mov eax,[ebx]  ; dword from table
         mov [edi],eax  ; into bitmap
         add edi,3
         inc esi        ; next index
         dec ecx
         jne @B         ; until line finished

         add edi,pad    ; add pad bytes to bitmap area
         add esi,Cpad   ; add pad bytes to the 8bit index
         dec edx
         jne more       ; new scan line

         jmp Bexit      ; end

;________________________________________________________________________________

ERROR:

     push eax
     invoke GlobalFree,hBmem
     invoke GlobalFree,hCol
     invoke GlobalFree,hBmem2
     pop eax
     stc
     ret

BmpFileTo24Bit endp

;________________________________________________________________________________
;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
; returns the padding for the color index data
;
; in  -> Plength = number of pixels wide
;     -> number of bits per pixel
;
; out <- carry clear  eax = size padding for one index line
;     <- carry set =  eax = 0  invalid bit count

ColorIxPad proc uses edx ebx Plength:dword,BitCount:dword

         mov eax,Plength

     .if BitCount == 1
         mov edx,eax
         and edx,7      ; div/8 remainder
         shr eax,3      ; quotient
         cmp edx,0
         je @F
         inc eax        ; line ends mid byte
     @@: and eax,3      ; div/4 remainder
         cmp eax,0
         je @F
         mov edx,4
         sub edx,eax    ; to make dword boudary
         mov eax,edx    ; pad bytes
     @@: clc
;        -----------------
     .elseif BitCount == 4
         shr eax,1
         jnc @F
         inc eax
     @@: and eax,3
         cmp eax,0
         je @F
         mov edx,4
         sub edx,eax
         mov eax,edx
     @@: clc
;        -----------------
     .elseif BitCount == 8
         and eax,3
         cmp eax,0
         je @F
         mov edx,4
         sub edx,eax
         mov eax,edx
     @@: clc
     .else
         mov eax,0
         stc
     .endif

ret
ColorIxPad endp

;________________________________________________________________________________
;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
; -- Returns the bitmap size including the pad bytes and the pad bytes

; in  -> Width, Height,ptr for PadBytes
; out <- eax = bitmapsize     
;     <- PadBytes = padding for one line

; if Pads   = 0 then eax = ((Bwidth*3) * Bheight)
; if Pads   = 1 then eax = (((Bwidth*3)+Padding) * Bheight)
; else Pads = Padding,  and  eax = (((Bwidth*3)+Padding) * Bheight)


Bitmapsize proc uses ebx edx Bwidth:dword,Bheight:dword,Pads:dword
local blength:dword

         invoke IntMul,Bwidth,3
         mov blength,eax
         and eax,3           ; remainder after div/4
         cmp eax,0
         je @F               ; 0 pad bytes
         mov ebx,4
         sub ebx,eax         
         mov eax,ebx         ; eax = bytes to pad out to DWORD boundary
     @@: cmp Pads,0
         jne @F              ; if Pads = 0 don't count pad bytes
         mov eax,0
         jmp ovr
     @@: cmp Pads,1          ; if pads = 1 count pad bytes (no Pads output)
         je ovr
         mov ebx,Pads
         mov [ebx],eax       ; else write number of pad bytes per line to Pads
    ovr: add eax,blength
         invoke IntMul,eax,Bheight
   ret
Bitmapsize endp
;________________________________________________________________________________
;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
;
;
;